#include <stdio.h>
#include <stdlib.h>
//#include <cuda_runtime.h>
#include "cuda-helper.h"
#include "cpu-helper.h"
#include "timing418.h"
#include "conv-kernel.h"

// Product must be smaller than 1024.
#define BLOCK_SIZE_X 32
#define BLOCK_SIZE_Y 32

// Product should be smaller than 10KB / 4.
#define MASK_MAX_P 41
#define MASK_MAX_Q 41

// Product should be smaller than 48KB / 4.
#define TILE_SIZE_X (BLOCK_SIZE_X + MASK_MAX_P - 1)
#define TILE_SIZE_Y (BLOCK_SIZE_Y + MASK_MAX_Q - 1)

// Declare constant memory for the mask array.
__constant__ float Mask[MASK_MAX_P * MASK_MAX_Q];
// Storage of the actual mask size.  Runtime doesn't appear to let me
// declare these as plain variables rather than arrays.
__constant__ uint P[1], Q[1], P_half[1], Q_half[1];


//----------------------------------------------------------------------
// We will use these values when confirming the optimal throughput of
// your tiled kernel.

// Must return a value between 500 and 15000.  For sample solution
// implementation it turns out that 1024 also gives about the same
// throughput, but throughput drops dramatically for 4096 and slightly
// for 512.
uint best_mn() {
  return 2048;
}

// Must return a value between 5 and 21.  For sample solution
// implementation, larger appears to be better.  There is a slight
// degradation at 11, more significant beyond that.  A value of 41
// achieved higher throughput, although that wasn't allowed.
uint best_pq() {
  return 21;
}

//----------------------------------------------------------------------
/* Copy the mask data into constant memory on the GPU.  Also record
   the mask's p and q for later use.
*/
void set_up_mask(const float *mask, const uint p, const uint q) {

  // Check that the mask is not too big.
  if((p > MASK_MAX_P) || (q > MASK_MAX_Q)) {
    fprintf(stderr, "\nCannot handle mask of size %d x %d (max %d x %d)",
            p, q, MASK_MAX_P, MASK_MAX_Q);
    exit(-1);
  }

  // Integer divide causes round down assuming precondition is met.
  const uint p_half = p / 2;
  const uint q_half = q / 2;

  // Copy over the mask parameters.

  cudaTry(cudaMemcpyToSymbol(P, &p, sizeof(uint)));
  cudaTry(cudaMemcpyToSymbol(Q, &q, sizeof(uint)));
  cudaTry(cudaMemcpyToSymbol(P_half, &p_half, sizeof(uint)));
  cudaTry(cudaMemcpyToSymbol(Q_half, &q_half, sizeof(uint)));

  cudaTry(cudaMemcpyToSymbol(Mask, mask, p*q*sizeof(float)));
}

//----------------------------------------------------------------------
/* Finished with this mask array.  Don't actually need to do anything
   because all of the memory is allocated statically.
*/
void clean_up_mask() {
  // No work to be done.
}


//----------------------------------------------------------------------
__global__ void conv_basic_kernel(const float *d_data_in, float *d_data_out,
                                  const uint m, const uint n) {

  // We usually interpret "x" as running horizontally, but in this
  // case we want the threads in a single warp to access sequential
  // memory locations, which for column major means sequential rows,
  // and that means that "x" has to refer to rows.
  const uint i = blockIdx.x * blockDim.x + threadIdx.x;
  const uint j = blockIdx.y * blockDim.y + threadIdx.y;

  if((i < m) && (j < n)) {
    // If this is a valid input data element.
    float sum = 0.0f;
    for(uint l = 0; l < Q[0]; l++) {
      // Which column of input data are we examining?
      const int j_offset = j - Q_half[0] + l;
      if((j_offset >= 0) && (j_offset < n))
        // If this is a valid column.
        for(uint k = 0; k < P[0]; k++) {
          // Which row of input data are we examining?
          const int i_offset = i - P_half[0] + k;
          if((i_offset >= 0) && (i_offset < m))
            // If this is a valid row.
            sum += (d_data_in[IDX2F(i_offset,j_offset,m,n)]
                    * Mask[IDX2F(k,l,P[0],Q[0])]);
        }
    }
    // Save the result.
    d_data_out[IDX2F(i,j,m,n)] = sum;
  }
}


//----------------------------------------------------------------------
__global__ void conv_basic_tiled(const float *d_data_in, float *d_data_out,
                                 const uint m, const uint n) {

  // We usually interpret "x" as running horizontally, but in this
  // case we want the threads in a single warp to access sequential
  // memory locations, which for column major means sequential rows,
  // and that means that "x" has to refer to rows.
  const uint i = blockIdx.x * blockDim.x + threadIdx.x;
  const uint j = blockIdx.y * blockDim.y + threadIdx.y;

  const uint it = threadIdx.x;
  const uint jt = threadIdx.y;

  // IT would be notationally simpler to use a C-style 2D array, but
  // that would be row-major ordering and we want to be consistent
  // about column-major so that threads in the same warp are accessing
  // consecutive shared memory locations and hence avoid bank conflicts.
  __shared__ float data_in_sh[TILE_SIZE_X * TILE_SIZE_Y];

  // Then get the target range of the halo
  const int halo_i_start = blockDim.x * blockIdx.x - P_half[0];
  const int halo_i_end = blockDim.x * (blockIdx.x + 1) + P_half[0];

  const int halo_j_start = blockDim.y * blockIdx.y - Q_half[0];
  const int halo_j_end = blockDim.y * (blockIdx.y + 1) + Q_half[0];
  
  // Load tile one block-size at a time.
  for(int l = jt; l + halo_j_start < halo_j_end; l += BLOCK_SIZE_Y) {
    const int j_offset = l + halo_j_start;
    for(int k = it; k + halo_i_start < halo_i_end; k += BLOCK_SIZE_X) {
      const int i_offset = k + halo_i_start;
      if((i_offset >= 0) && (i_offset < m) 
         && (j_offset >= 0) && (j_offset < n))
        // This is a valid input data item.
        data_in_sh[IDX2F(k,l,TILE_SIZE_X,TILE_SIZE_Y)] = 
          d_data_in[IDX2F(i_offset,j_offset,m,n)];
      else
        // This data item lies outside the valid data, so use boundary
        // condition rule (value 0.0).
        data_in_sh[IDX2F(k,l,TILE_SIZE_X,TILE_SIZE_Y)] = 0.f;
    }
  }

  __syncthreads();
  
  if((i < m) && (j < n)) {
    // If this is a valid input data element.
    float sum = 0.0f;
    for(uint l = 0; l < Q[0]; l++) {
      // Which column of input data are we examining?
      const int j_offset = jt + l;
      for(uint k = 0; k < P[0]; k++) {
        // Which row of input data are we examining?
        const int i_offset = it + k;
        sum += (data_in_sh[IDX2F(i_offset,j_offset,TILE_SIZE_X,TILE_SIZE_Y)]
                * Mask[IDX2F(k,l,P[0],Q[0])]);
       }
    }
    // Save the result.
    d_data_out[IDX2F(i,j,m,n)] = sum;
  }
}


//----------------------------------------------------------------------
// Computes the convolution on the GPU in the very basic manner: One
// thread per output pixel with all data read or written directly to
// global memory.
void conv_gpu(const float *d_data_in, float *d_data_out,
              const uint m, const uint n, const bool tiled) {

  // We will only use a single block.
  dim3 dimGrid(ceil(m / (double)BLOCK_SIZE_X), 
               ceil(n / (double)BLOCK_SIZE_Y), 1);
  dim3 dimBlock(BLOCK_SIZE_X, BLOCK_SIZE_Y, 1);

  if(tiled)
    conv_basic_tiled<<<dimGrid, dimBlock>>>(d_data_in, d_data_out, m, n);
  else
    conv_basic_kernel<<<dimGrid, dimBlock>>>(d_data_in, d_data_out, m, n);
}
