//#include <stdio.h>
//#include <stdlib.h>
//#include <cuda_runtime.h>
//#include "cuda-helper.h"
//#include "cpu-helper.h"
//#include "timing418.h"
#include "moments-kernel.h"

/* We use a relatively easy to implement version of reduce: Use just
   one block of maximum size.  In the first phase all threads divide
   the full data set into chunks of appropriate size and perform
   independent reductions with results written to shared memory.  In
   the second phase a reduce tree is constructed in shared memory
   following the improved reduce pattern from K&H figure 5.15.

   Better results could likely be accomplished with a multi-phase
   reduction with multiple blocks in the first phase, as discussed
   near the end of K&H chapter 8 for scan.
*/

#define BLOCK_SIZE 1024

//----------------------------------------------------------------------
// The functions to provide what I measured as the best choices for
// the parameters.

uint best_n() {
  return 100000000;
}

uint best_moment() {
  return 2;
}

//----------------------------------------------------------------------
// Note that mean is stored into a memory location on the GPU, rather
// than being returned.
__global__ void compute_mean_kernel(const float *x, const uint n,
                                    double *mean) {

  // Implementation assumes just one block, so ignore the block index.
  uint i = threadIdx.x;
  
  // Allocate shared memory to hold the partial sum for this thread.
  __shared__ double sums[BLOCK_SIZE];
  // Set all sums to zero since the reduction tree will add elements
  // beyond n if n < BLOCK_SIZE.
  sums[i] = 0.0;

  // Compute partial sums for each thread independently.
  for(uint j = i; j < n; j += BLOCK_SIZE)
    sums[i] += x[j];

  // Perform the reduction tree.
  for(uint stride = BLOCK_SIZE >> 1; stride >= 1; stride = stride >> 1) {
    __syncthreads();
    if(i < stride)
      sums[i] += sums[i + stride];
  }

  // Copy the final result out of shared memory and back to global memory.
  if(i == 0)
    mean[0] = sums[0] / n;
}

//----------------------------------------------------------------------
// Returns all moments from first up to max_moment.  Note that the
// first moment (index 0) will be set to the mean.
__global__ void compute_moments_kernel(const float *x, const uint n,
                                       const double *mean_in,
                                       const uint max_moment,
                                       double *moments) {

  // Implementation assumes just one block, so ignore the block index.
  uint i = threadIdx.x;

  // Copy the mean value out of global memory and into a local variable.
  double mean = mean_in[0];
  
  // Allocate shared memory to hold the partial sums for this thread.
  // Note that we do not allocate space for the first moment, so we
  // have to be careful not to index that first moment.  We also
  // allocate all k^th moments together to remove bank conflicts.
  __shared__ double sums[BLOCK_SIZE * (MAX_MOMENT - 1)];
  for(uint k = 0; k < max_moment - 1; k++)
    sums[i + k * BLOCK_SIZE] = 0.0;

  // Compute partial sums for each thread independently.
  for(uint j = i; j < n; j += BLOCK_SIZE) {
    double offset = x[j] - mean;
    double product = offset;
    for(uint k = 0; k < max_moment - 1; k++) {
      product *= offset;
      sums[i + k * BLOCK_SIZE] += product;
    }
  }

  // Perform the reduction tree.
  for(uint stride = BLOCK_SIZE >> 1; stride >= 1; stride = stride >> 1) {
    __syncthreads();
    if(i < stride)
      for(uint k = 0; k < max_moment - 1; k++)
        sums[i + k * BLOCK_SIZE] += sums[i + k * BLOCK_SIZE + stride];
  }

  // Need to synchronize again because we are going to spread the
  // final division operation across the threads.
  __syncthreads();

  if(i == 0)
    // Copy over the mean value.
    moments[0] = mean;
  else if(i < max_moment)
    // Perform the division and copy the result out to global memory
    // for the remaining moments.  Remember about the indexing of
    // moments in shared memory.
    moments[i] = sums[(i-1) * BLOCK_SIZE] / n;
}

//----------------------------------------------------------------------
// Bundles the computation of the mean and moments.  Note that arrays
// d_x and d_moments are assumed to be allocated on the GPU already.
void compute_all_gpu(const float *d_x, const uint n,
                     const uint max_moment, double *d_moments) {

  // We will only use a single block.
  dim3 dimGrid(1, 1, 1);
  dim3 dimBlock(BLOCK_SIZE, 1, 1);

  // Calculate the mean.  Note that the final argument only needs
  // enough room to store one double (the mean), but we pass in the
  // entire moments array for convenience.
  compute_mean_kernel<<<dimGrid, dimBlock>>>(d_x, n, d_moments);

  // Apparently we don't need to explicitly wait for the kernel to
  // finish, since kernels will by default be placed into the same
  // stream and kernels in the same stream execute sequentially.
  // cudaDeviceSynchronize();  

  // Calculate the other moments.
  compute_moments_kernel<<<dimGrid, dimBlock>>>(d_x, n, d_moments, 
                                                max_moment, d_moments);
}
