#include <stdio.h>
//#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
//#include <assert.h>
#include "timing418.h"
#include "cpu-helper.h"
#include "moments-loader.h"
#include "moments-kernel.h"

#define CLOSE_ENOUGH 1e-5

//--------------------------------------------------------------------------
// Random floats between lower and upper (inclusive).  Note that we
// construct the number using double precision.
void rand_float_vector(float *x, const uint n, 
                       const double lower, const double upper) {

  double gap = upper - lower;
  
  for(uint i = 0; i < n; i++) {
    double unit_uniform = (double)rand() / (double)RAND_MAX;
    x[i] = (float)(gap * unit_uniform + lower);
  }
}

//--------------------------------------------------------------------------
// Sequence of float values.
void seq_float_vector(float *x, const uint n, 
                      const double start, const double step) {

  double value = start;
  for(uint i = 0; i < n; i++) {
    x[i] = value;
    value += step;
  }
}

//----------------------------------------------------------------------
void check_double_vectors(const double *x1, const double *x2,
                         const uint n, const char *s1, const char *s2) {

  for(uint i = 0; i < n; i++) {
    if(fabs(x1[i] - x2[i]) > CLOSE_ENOUGH) {
      fprintf(stderr, "\nERROR: element=%d, %s=%f, %s=%f\n",
              i, s1, x1[i], s2, x2[i]);
      exit(-1);
    }
    if(isnan(x1[i]) || isnan(x2[i])) {
      fprintf(stderr, "\nNaN detected: element=%d, %s=%f, %s=%f\n",
              i, s1, x1[i], s2, x2[i]);
      exit(-1);
    }
  }
}

//----------------------------------------------------------------------
double compute_mean_cpu(const float *x, const uint n) {

  double sum = 0.0;
  for(uint i = 0; i < n; i++)
    sum += x[i];
  return sum / n;
}

//----------------------------------------------------------------------
// Returns all moments from first up to max_moment.  Note that the
// first moment (index 0) will be set to the mean.
void compute_moments_cpu(const float *x, const uint n,
                         const double mean, const uint max_moment,
                         double *moments) {
  // We will compute partial sums in the moments array, so we need to
  // clear it.
  for(uint k = 1; k < max_moment; k++)
    moments[k] = 0.0;
  for(uint i = 0; i < n; i++) {
    double offset = x[i] - mean;
    double product = offset;
    for(uint k = 1; k < max_moment; k++) {
      product *= offset;
      moments[k] += product;
    }
  }
  
  for(uint k = 1; k < max_moment; k++)
    moments[k] /= n;

  moments[0] = mean;
}

    
//----------------------------------------------------------------------
// Bundles the computation of the mean and moments.
void compute_all_cpu(const float *x, const uint n,
                     const uint max_moment, double *moments) {
  double mean_cpu = compute_mean_cpu(x, n);
  compute_moments_cpu(x, n, mean_cpu, max_moment, moments);
}

//----------------------------------------------------------------------
// Compute the FP operations for computing all the moments of a
// sequence from first up to max_moment.  Assume n >> 1 so that final
// divide by n doesn't matter.
uint total_fp_ops(const uint n, const uint max_moment) {

  // Each moment requires a product and a sum for each sample.  Note
  // that the first moment's cost is computing the mean.
  return (n * 2 * max_moment);
}

//----------------------------------------------------------------------
void moments_protocol(const uint n, const uint max_moment,
                      const double lower, const double upper) {

  // Always use odd number of trials so that median is easy to calculate.
  const uint trials = 5;
  timer_418 t0, t1;

  // Place to store the data.
  float *x;
  double *mom_cpu, *mom_gpu;
  double *run_times;

  // Allocate the vectors.
  NewArrayCPU(float, n, x);
  NewArrayCPU(double, max_moment, mom_cpu);
  NewArrayCPU(double, max_moment, mom_gpu);
  NewArrayCPU(double, trials, run_times);

  // Fill the initial vector.
  rand_float_vector(x, n, lower, upper);
  //seq_float_vector(x, n, lower, (upper - lower) / (n - 1));

  /*
  // Print the input data.  For debugging.
  printf("\nInput data (%d elements): [ %5.3e", n, x[0]);
  for(uint i = 1; i < n; i++)
    printf(", %5.3e", x[i]);
  printf(" ]\n");
  */

  // Compute the moments.  Time the runs and report median times.  Check that
  // the final results agree (a minimal test).
  printf("\n\nMoments up to %d on CPU: ", max_moment);
  for(uint i = 0; i < trials; i++) {
    timestamp(&t0);
    compute_all_cpu(x, n, max_moment, mom_cpu);
    timestamp(&t1);
    run_times[i] = elapsed_seconds(t1, t0);
    printf(" %5.2e", run_times[i]);
  }
  double median_time_cpu = median_time(run_times, trials);
  printf(" -- median %5.2e, flops %5.2e",
         median_time_cpu, total_fp_ops(n, max_moment) / median_time_cpu);

  printf("\n\nMoments up to %d on GPU: ", max_moment);
  // Pointers to the data on the GPU.
  float *d_x;
  double *d_moments;
  data_to_gpu(x, n, max_moment, &d_x, &d_moments);
  for(uint i = 0; i < trials; i++) {
    timestamp(&t0);
    compute_all_gpu(d_x, n, max_moment, d_moments);
    after_run_gpu();
    timestamp(&t1);
    run_times[i] = elapsed_seconds(t1, t0);
    printf(" %5.2e", run_times[i]);
  }
  data_from_gpu(mom_gpu, max_moment, d_x, d_moments);
  double median_time_gpu = median_time(run_times, trials);
  printf(" -- median %5.2e, flops %5.2e",
         median_time_gpu, total_fp_ops(n, max_moment) / median_time_gpu);
  //check_double_vectors(mom_cpu, mom_gpu, n, "CPU", "GPU");

  // Show the results.
  for(uint j = 0; j < max_moment; j++) {
    printf("\nMoment %d: [CPU] %f, [GPU] %f", j+1, mom_cpu[j], mom_gpu[j]);
  }

  // Free up all the CPU memory resources.
  free(x);
  free(mom_cpu);
  free(mom_gpu);
  free(run_times);
}


//----------------------------------------------------------------------
int main(int argc, char **argv) {

  uint max_moment_default = 2;
  double lower_default = 0.0f, upper_default = 1.0f; 
  uint n, max_moment;
  double lower, upper;
  char *conversion_garbage;

  if((argc < 2) || (argc > 5)) {
    fprintf(stderr, "usage: moments n [max_moment = %d] [lower = %4.1f]  [upper = %4.1f]\n",
            max_moment_default, lower_default, upper_default);
    exit(1);
  }

  // Parse the input arguments.
  n = strtoul(argv[1], &conversion_garbage, 10);
  if(strlen(conversion_garbage) > 0) {
    fprintf(stderr, "moments unable to convert %s to uint n\n", argv[1]);
    exit(1);
  }

  if(argc > 2) {
    max_moment = strtoul(argv[2], &conversion_garbage, 10);
    if((strlen(conversion_garbage) > 0) 
       || (max_moment < 2) || (max_moment > MAX_MOMENT)) {
      fprintf(stderr, "moments unable to convert %s to uint max_moment in range [ 2, %d ]\n", argv[2], MAX_MOMENT);
      exit(1);
    }
  }
  else
    max_moment = max_moment_default;

  if(argc > 3) {
    lower = strtof(argv[3], &conversion_garbage);
    if(strlen(conversion_garbage) > 0) {
      fprintf(stderr, "moments unable to convert %s to double lower\n", argv[3]);
      exit(1);
    }
  }
  else
    lower = lower_default;

  if(argc > 3) {
    upper = strtof(argv[4], &conversion_garbage);
    if((lower >= upper) || (strlen(conversion_garbage) > 0)) {
      fprintf(stderr, "moments unable to convert %s to double upper > %f\n", argv[4], lower);
      exit(1);
    }
  }
  else
    upper = upper_default;

  // Uncomment the call to srand() if you want different random number
  // sequences on each run.
  //srand((uint)time(NULL));

  printf("Computing %d moments for %d samples in range [ %f, %f ]",
         max_moment, n, lower, upper);
  moments_protocol(n, max_moment, lower, upper);
  printf("\n\n");
}
